/*----------------------------------------------------------------------------
 * mapgen.c	- translate map tables (version 2)
 *
 * usage: mapgen file ...
 *
 * Directory:
 * main (argc, argv)				- check correct usage
 * translate (fp_in, fp_out)			- translate a map file
 * translate_keymap (fp_in)			- translate key map part
 * search_key (kname)				- search a key name
 * convert_sequence (target, sequence, maxlen)	- convert sequence to binary
 *
 * Compilation: cc -O -I/usr/local/include mapgen.c -o mapgen
 *
 *----------------------------------------------------------------------------
 */

#include <ecurses.h>			/* ecurses.h in /usr/local/include  */

#include "map-table.h"				    /* data structures	    */

extern void exit ();

/*----------------------------------------------------------------------------
 * main (argc, argv)				- check correct usage
 *----------------------------------------------------------------------------
 */

int
main (argc, argv)
int argc;
char * argv[];
{
    FILE * fp_in;			    /* file pointer for read	    */
    FILE * fp_out;			    /* file pointer for write	    */
    char fname[128];			    /* new filename		    */
    int length;				    /* length of filename	    */
    int rtc;				    /* return code of translate()   */
    int i;				    /* counter			    */

    if (argc < 2)
    {
	fprintf (stderr, "usage: %s file ...\n", argv[0]);
	exit (1);
    }

    for (i = 1; i < argc; i++)		    /* loop over all files ...	    */
    {
	strcpy (fname, argv[i]);	    /* check correct suffix	    */

	length = strlen (fname);

	if (length < 4 || strcmp (fname + length - 3, ".IN"))
	{
	    fprintf (stderr, "%s: %s has no suffix '.IN' !\n", argv[0], fname);
	    exit (1);
	}

	fp_in = fopen (fname, "r");			/* open file	    */

	if (! fp_in)					/* successful ?	    */
	{						/* no !		    */
	    fprintf (stderr, "%s: cannot open %s !\n", argv[0], fname);
	    exit (1);
	}

	strcpy (fname + length - 3, ".OUT");		/* xx.IN --> xx.OUT */

	fp_out = fopen (fname, "w");			/* open output file */

	if (! fp_out)
	{						/* open error !	    */
	    fprintf (stderr, "%s: cannot open %s !\n", argv[0], fname);
	    fclose (fp_in);
	    exit (1);
	}

	rtc = translate (fp_in, fp_out);		/* translate table  */

	(void) fclose (fp_in);				/* close the files  */
	(void) fclose (fp_out);

	if (rtc != 0)
	{
	    exit (1);					/* error occured !  */
	}
	printf ("%s --> %s\n", argv[i], fname);		/* information	    */
    }

    return (0);
} /* main (argc, argv[]) */


/*----------------------------------------------------------------------------
 * translate (fp_in, fp_out)			- translate a map file
 *----------------------------------------------------------------------------
 */

int
translate (fp_in, fp_out)
FILE * fp_in;
FILE * fp_out;
{
    char buf[256];					/* input buffer	    */
    char cname[256];					/* character name   */
    unsigned char esc_sequence[64];			/* input sequence   */
    int i, j;						/* counters	    */
    int chset_no;					/* character set no */
    int iso_char;					/* iso character    */
    int ch_out;						/* output character */
    int in_first_part = TRUE;				/* flag: 1st or 2nd */
    int rtc;						/* return code	    */

    for (i = 0; i < MAX_USED_CHARSETS; i++)	/* 1st clear charset array! */
    {						/* because this function    */
	for (j = 0; j < MAX_ESC_LEN; j++)	/* could be called more	    */
	{					/* than once		    */
	    charset[i][j] = '\0';
	}
    }

    for (i = 0; i < 256; i++)			/* clear also map_table	    */
    {
	map_table[i].chset_no = 0;		/* default: ASCII chset #0  */
	map_table[i].ch_out = i;		/* default: no translation  */
    }

    while (fgets (buf, 255, fp_in))		/* read line by line ..	    */
    {
	if (*buf != '#' && *buf != '\n')	/* Comment or empty line ?  */
	{					/* no ...		    */
	    if (in_first_part)			/* in 1st part of file ?    */
	    {					/* yes			    */
		if (! strncmp (buf, "END", 3))	/* end of 1st part ?	    */
		{				/* yes, reset flag	    */
		    in_first_part = FALSE;
		}
		else				/* read charset number and  */ 
		{				/* escape sequence	    */
		    if (sscanf (buf, "%d %s", &chset_no, esc_sequence) != 2)
		    {
			fprintf (stderr, "Read error.\n");
			return (-1);
		    }
		    if (chset_no >= MAX_USED_CHARSETS)
		    {
			fprintf (stderr, "Illegal character set number.\n");
			return (-1);
		    }

		    if (convert_sequence (charset[chset_no], esc_sequence,
			MAX_ESC_LEN) == -1)
		    {
			return (-1);
		    }
		}
	    }
	    else				/* the 2nd part ...	    */
	    {
		if (! strncmp (buf, "KEYBOARD", 8))
		{				    /* keymap follows !	    */
		    rtc = translate_keymap (fp_in);

		    if (rtc != 0)
		    {
			return (rtc);
		    }
		    break;			/* we are ready !	    */
		}
		if (sscanf (buf, "%s %d %x",	/* read three columns	    */
			    cname, &chset_no, &ch_out) != 3)
		{
		    fprintf (stderr, "Read error.\n");
		    return (-1);
		}

		if ((iso_char = search_key (cname)) == -1)
		{
		    return (-1);
		}

		map_table[iso_char].chset_no = chset_no;    /* store chset# */
		map_table[iso_char].ch_out = ch_out;	    /* and char	    */
	    }
	}
    }

    fwrite (charset, sizeof (charset), 1, fp_out);	/* write charset    */
    fwrite (map_table, sizeof (map_table), 1, fp_out);	/* write map_table  */
    fwrite (key_table, sizeof (key_table), 1, fp_out);	/* write key_table  */

    return (0);						/* keymap table	    */
} /* translate (fp_in, fp_out) */


/*----------------------------------------------------------------------------
 * translate_keymap (fp_in)			- translate key map part
 *----------------------------------------------------------------------------
 */

int
translate_keymap (fp_in)
FILE * fp_in;
{
    char buf[256];					/* input buffer	    */
    int scan;						/* scan-code	    */
    char name_buf[64];					/* name of key	    */
    unsigned char sequence_buf[64];			/* sequence	    */

    for (scan = 0; scan < 256; scan++)		/* 1st set default values   */
    {						/* of alphanumeric keys	    */
	key_table[scan][0] = scan;
	key_table[scan][1] = '\0';
    }
    key_table[0x1b][1] = 0x1b;			/* double press escape	    */
    key_table[0x1b][2] = '\0';

    for (scan = 256; scan < 512; scan++)	/* 2nd clear all func keys  */
    {
	key_table[scan][0] = '\0';
    }

    while (fgets (buf, 255, fp_in))		/* read line by line ..	    */
    {
	if (*buf != '#' && *buf != '\n')	/* Comment or empty line ?  */
	{					/* no ...		    */
	    if (sscanf (buf, "%s %s", name_buf, sequence_buf) != 2)
	    {
		fprintf (stderr, "Read error.\n");
		return (-1);
	    }

	    if ((scan = search_key (name_buf)) == -1)	/* search key name  */
	    {
		fprintf (stderr, "Unknown key name: %s\n", name_buf);
		return (-1);
	    }

	    if (convert_sequence (key_table[scan], sequence_buf,
		MAX_KEY_SEQ_LEN) == -1)
	    {
		return (-1);
	    }
	}
    }
    return (0);						/* return: success  */
} /* translate_keymap (fp_in) */


/*----------------------------------------------------------------------------
 * search_key (kname)				- search a key name
 *----------------------------------------------------------------------------
 */

int
search_key (kname)
char * kname;
{
    char compare_buffer[16];			    /* compare buffer	    */
    int scan;					    /* scan code	    */

    for (scan = 0; keyname_table[scan]; scan++)	    /* 1st search in	    */
    {						    /* keyname_table[]	    */
	if (! strcmp (kname, keyname_table[scan]))
	{
	    return (scan);			    /* found, return scan   */
	}
    }

    for (scan = 0; scan < 32; scan++)		    /* look for Function    */
    {						    /* key name ...	    */

	sprintf (compare_buffer, "KEY_F%d", scan);	/* NORMAL Fkeys	    */
	if (! strcmp (kname, compare_buffer))
	{
	    return (scan + KEY_F0);
	}

	sprintf (compare_buffer, "KEY_SF%d", scan);	/* SHIFT Fkeys	    */
	if (! strcmp (kname, compare_buffer))
	{
	    return (scan + KEY_SF0);
	}

	sprintf (compare_buffer, "KEY_CF%d", scan);	/* CTRL Fkeys	    */
	if (! strcmp (kname, compare_buffer))
	{
	    return (scan + KEY_CF0);
	}

	sprintf (compare_buffer, "KEY_AF%d", scan);	/* ALT Fkeys	    */
	if (! strcmp (kname, compare_buffer))
	{
	    return (scan + KEY_AF0);
	}

	sprintf (compare_buffer, "KEY_SCF%d", scan);	/* SHIFT CTRL Fkeys */
	if (! strcmp (kname, compare_buffer))
	{
	    return (scan + KEY_SCF0);
	}

	sprintf (compare_buffer, "KEY_SAF%d", scan);	/* SHIFT ALT Fkeys  */
	if (! strcmp (kname, compare_buffer))
	{
	    return (scan + KEY_SAF0);
	}
    }

    return (-1);					/* key not found    */
} /* search_key (kname) */


/*----------------------------------------------------------------------------
 * convert_sequence (target, sequence, maxlen)	- convert sequence to binary
 *----------------------------------------------------------------------------
 */

int
convert_sequence (target, sequence, maxlen)
unsigned char * target;
unsigned char * sequence;
int maxlen;
{
    static char	hex_buf[3];			/* static ! (strncpy below) */
    int		hex_code;			/* hex code of hex_buf	    */
    int		len = 0;			/* current length	    */
    unsigned char * seq_ptr = sequence;		/* pointer		    */

    while (* seq_ptr)
    {
	if (* seq_ptr == '\\')
	{
	    seq_ptr++;

	    switch (* seq_ptr)
	    {
		case '\\':
		    *target = '\\';
		    break;

		case 'E':
		case 'e':
		    *target = 0x1b;
		    break;

		case 'X':
		case 'x':
		    strncpy (hex_buf, (char *) seq_ptr, 2);

		    if (sscanf (hex_buf, "%02x", &hex_code) != 1)
		    {
			fprintf (stderr, "Illegal HEX-CODE: %s\n", hex_buf);
			return (-1);
		    }

		    *target = hex_code;
		    seq_ptr++;
		    break;

		default:
		    *target = *seq_ptr;
	    }
	}
	else
	{
	    *target = *seq_ptr;
	}

	len++;

	if (len == maxlen)
	{
	    fprintf (stderr, "Too long sequence: %s\n", sequence);
	    return (-1);
	}

	seq_ptr++;
	target++;
    }
    *target = '\0';

    return (0);
} /* convert_sequence (target, sequence, maxlen) */

/* end of mapgen.c */
